#![feature(slice_as_chunks)]

pub mod mmap;

use std::fmt::Display;

#[derive(Copy, Clone)]
pub struct StationMeasurements {
    pub min: isize,
    pub max: isize,
    pub count: isize,
    pub sum: isize,
}

impl StationMeasurements {
    pub fn update(&mut self, v: isize) {
        self.min = self.min.min(v);
        self.max = self.max.max(v);
        self.count += 1;
        self.sum += v;
    }

    pub fn merge(&mut self, other: &Self) {
        self.min = self.min.min(other.min);
        self.max = self.max.max(other.max);
        self.count += other.count;
        self.sum += other.sum;
    }
}

impl Display for StationMeasurements {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let min = self.min as f64 / 10.0;
        let max = self.max as f64 / 10.0;
        let avg = (self.sum as f64 / self.count as f64) / 10.0;
        write!(f, "{min}/{avg:.1}/{max}")
    }
}

pub fn format_nums(num: usize) -> String {
    num.to_string()
        .as_bytes()
        .rchunks(3)
        .rev()
        .map(std::str::from_utf8)
        .collect::<Result<Vec<&str>, _>>()
        .unwrap()
        .join("_")
}

#[inline]
pub const fn get_digit(b: u8) -> isize {
    // wrapping_sub('0' as u32) same as - 48 but less magical
    (b as isize).wrapping_sub('0' as isize)
}

#[inline]
pub fn parse_temp(bytes: &[u8]) -> isize {
    let is_negative = bytes[0] == b'-';
    let as_decimal = match (is_negative, bytes.len()) {
        (true, 4) => get_digit(bytes[1]) * 10 + get_digit(bytes[3]),
        (true, 5) => get_digit(bytes[1]) * 100 + get_digit(bytes[2]) * 10 + get_digit(bytes[4]),
        (false, 3) => get_digit(bytes[0]) * 10 + get_digit(bytes[2]),
        (false, 4) => get_digit(bytes[0]) * 100 + get_digit(bytes[1]) * 10 + get_digit(bytes[3]),
        _x => panic!("could not parse temp: is_negative = {is_negative}, length = {}", bytes.len()),
    };
    if is_negative {
        -as_decimal
    } else {
        as_decimal
    }
}

#[inline]
pub fn hashstr(bytes: &[u8]) -> usize {
    let mut hash = 0;
    let (chunks, remainder) = bytes.as_chunks::<8>();
    for &chunk in chunks {
        hash += usize::from_be_bytes(chunk);
    }
    let mut r = [0_u8; 8];
    r[0] = remainder.len() as u8;
    let mut idx = 1;
    for &byte in remainder {
        r[idx] = byte;
        idx += 1;
    }
    hash += usize::from_be_bytes(r);
    hash
}

// not really faster
#[inline]
pub fn new_parse_temp(bytes: &[u8]) -> isize {
    let (characteristic, mantissa) = unsafe { bytes.split_at_unchecked(bytes.len() - 2) };
    let mut sum = (mantissa[1] as u32).wrapping_sub('0' as u32);
    let mut position = 10;
    let mut idx = characteristic.len();
    while idx > 0 {
        idx -= 1;
        if characteristic[idx] == b'-' {
            return sum as isize * -1;
        }
        sum += (characteristic[idx] as u32).wrapping_sub('0' as u32) * position;
        position *= 10;
    }
    sum as isize
}

// using Bytes struct has more performance impact than the std read_until method which uses Vec instead of slice
// #[inline]
// pub fn read_bytes_until(bytes: &mut Bytes<BufReader<&File>>, delimiter: u8) -> Option<Vec<u8>> {
//     let mut buf: Vec<u8> = Vec::with_capacity(108);
//     for byte in bytes {
//         if byte.is_err() {
//             panic!("Could not read byte");
//         }
//         let byte = byte.unwrap();
//         if delimiter == byte {
//             return Some(buf);
//         }
//         buf.push(byte);
//     }
//     None
// }

#[inline]
pub fn get_pos(bytes: &[u8], find: u8) -> Option<u32> {
    let chunks = bytes.windows(4);
    let mut pos = 0;
    for chunk in chunks {
        let inner_pos = get_pos_in_chunk(chunk, find);
        if inner_pos < chunk.len() as u32 {
            return Some(pos + inner_pos);
        }
        pos += 1;
    }
    None
}

#[inline]
fn get_pos_in_chunk(byte_chunk: &[u8], find: u8) -> u32 {
    let find_hex = u32::from_be_bytes([find; 4]);
    let x = u32::from_be_bytes(byte_chunk.try_into().unwrap()) ^ find_hex;
    let mask = (x - 0x01010101) & (!x & (0x80808080));
    u32::leading_zeros(mask) >> 3
}

#[cfg(test)]
mod tests {
    use crate::{get_pos, hashstr, new_parse_temp};

    #[test]
    fn test_new_parse_temp_max() {
        let temp_max = new_parse_temp("99.9".as_bytes());
        assert_eq!(temp_max, 999);
    }

    #[test]
    fn test_new_parse_temp_min() {
        let temp_min = new_parse_temp("-99.9".as_bytes());
        assert_eq!(temp_min, -999);
    }

    #[test]
    fn test_new_parse_temp_zero() {
        let temp_0 = new_parse_temp("0.0".as_bytes());
        assert_eq!(temp_0, 0);
    }

    #[test]
    fn test_new_parse_temp_pos() {
        let temp_10 = new_parse_temp("9.9".as_bytes());
        assert_eq!(temp_10, 99);
    }

    #[test]
    fn test_new_parse_temp_neg() {
        let temp_neg_10 = new_parse_temp("-9.9".as_bytes());
        assert_eq!(temp_neg_10, -99);
    }

    #[test]
    fn test_hashstr() {
        let hash_1 = hashstr(b"abcdefghijk");
        let hash_2 = hashstr(b"kjihgfedcba");
        let hash_3 = hashstr(b"abba");
        let hash_4 = hashstr(b"baab");

        assert_ne!(hash_1, hash_2);
        assert_ne!(hash_3, hash_4);
    }

    #[test]
    fn test_getpos() {
        let semi_bytes = vec![0_u8, 0_u8, 0_u8, 0_u8, 0_u8, 0_u8, 0_u8, 0_u8, b';', 0_u8, 0_u8];
        let semi_bytes = semi_bytes.as_slice();
        let pos = get_pos(semi_bytes, b';').unwrap();
        assert_eq!(pos, 8);
    }

    #[test]
    fn test_getpos_empty() {
        let semi_bytes = vec![0_u8, 0_u8, 0_u8, 0_u8, 0_u8, 0_u8, 0_u8, 0_u8, 0_u8, 0_u8];
        let semi_bytes = semi_bytes.as_slice();
        let pos = get_pos(semi_bytes, b';');
        assert_eq!(pos, None);
    }
}
